!-------------------------------------------------------------------------------
! Description:
!
!   Transfer model data to RTTOV profiles. A list of index values for the model
!   fields specify which profiles to transfer.
!
! Copyright:
!
!   This software was developed within the context of the EUMETSAT Satellite
!   Application Facility on Numerical Weather Prediction (NWP SAF), under the
!   Cooperation Agreement dated 7 December 2016, between EUMETSAT and the
!   Met Office, UK, by one or more partners within the NWP SAF. The partners
!   in the NWP SAF are the Met Office, ECMWF, DWD and MeteoFrance.
!
!   Copyright 2018, EUMETSAT, All Rights Reserved.
!
!-------------------------------------------------------------------------------

subroutine radsim_model_to_rttov( &
  prof_list,   & ! in
  model,       & ! in
  obs,         & ! inout
  profiles,    & ! inout
  cld_profiles ) ! inout

use parkind1, only : &
  jpim

use rttov_const, only : &
  qmin, &
  o3min

use radsim_mod_cfg, only : &
  seaice_threshold, &
  coast_threshold, &
  run_scatt, &
  clw_data, &
  use_totalice, &
  obs_grid, &
  run_scatt, &
  mmr_snowrain, &
  ir_addclouds, &
  ircloud_idg, &
  ircloud_ice_scheme, &
  ozone_data, &
  zen_angle, &
  output_mode

use radsim_mod_constants, only : &
  real64, &
  output_debug, &
  qcinfo_coast

use radsim_mod_process, only : &
  z_direction

use radsim_mod_types, only : &
  model_type, &
  obs_type

use rttov_const, only : &
  surftype_land, &
  surftype_sea, &
  surftype_seaice

use rttov_types, only : &
  rttov_profile,      &
  rttov_profile_cloud

implicit none

integer,                                intent(in)    :: prof_list(:)
type(model_type),                       intent(in)    :: model
type(obs_type),                         intent(inout) :: obs
type(rttov_profile),                    intent(inout) :: profiles(:)
type(rttov_profile_cloud), allocatable, intent(inout) :: cld_profiles(:)

integer :: i, j
integer(kind=jpim) :: nprofs, nlevels, nhalflevels
integer :: level1, layer1, halflevel1
integer :: prof
real(real64) :: cloud, cfrac
integer :: lsm
real :: zsurf
integer :: offset
logical :: fill_top
real :: landfrac

!-------------------------------------------------------------------------------

nprofs = size(prof_list)
nlevels = profiles(1) % nlevels

!-------------------------------------------------------------------------------
! 1. Set date and time
!-------------------------------------------------------------------------------

if ( size(model % validity_time, dim=1) == 1 ) then
  do i = 1, nprofs
    profiles(i) % date      = model % validity_time(1,1:3)
    profiles(i) % time(1:2) = model % validity_time(1,4:5)
  end do
else
  do i = 1, nprofs
    prof = prof_list(i)
    profiles(i) % date      = model % validity_time(prof,1:3)
    profiles(i) % time(1:2) = model % validity_time(prof,4:5)
  end do
end if

!-------------------------------------------------------------------------------
! 2. Single-level variables
!-------------------------------------------------------------------------------

do i = 1, nprofs

  prof = prof_list(i)

  !-----------------
  ! 2.1 Surface skin
  !-----------------

  ! Surface type. The input LSM is technically a land fraction so take the
  ! nearest integer value. If available, the provided observation value
  ! is preferred over the model value which may have been interpolated.

  if ( associated(obs % lsm) ) then
    landfrac = obs % lsm(prof)
  else
    landfrac = model % lsm(prof)
  end if

  lsm = nint(landfrac)

  if ( lsm == 1 ) then
    profiles(i) % skin % surftype = surftype_land
  else
    profiles(i) % skin % surftype = surftype_sea
  end if

  ! Set sea points to ice if seaice is over the defined threshold

  if ( associated(model % seaice) ) then
    if ( profiles(i) % skin % surftype == surftype_sea .and. &
         model % seaice(prof) > seaice_threshold                ) then
      profiles(i) % skin % surftype = surftype_seaice
    end if
  end if

  ! Flag point as coastline if the difference between the derived surface type
  ! and the model land fraction is over the defined threshold.

  landfrac = model % lsm(prof)
  if ( abs(landfrac-lsm) > coast_threshold ) then
    obs % qcinfo(prof) = ibset(obs % qcinfo(prof), qcinfo_coast)
  end if

  profiles(i) % skin % watertype = 1 ! 1=Ocean, 0=Freshwater
  profiles(i) % skin % t = model % tskin(prof)

  ! Set salinity based on watertype

  if ( profiles(i) % skin % watertype == 1 ) then
    profiles(i) % skin % salinity = 35.
  else
    profiles(i) % skin % salinity = 0.
  endif

  ! Fastem parameters for land/sea-ice are not used as RadSim specifies
  ! fixed land/sea-ice emissivity values.

  !---------------
  ! 2.2 Surface 2m
  !---------------

  profiles(i) % s2m % t     = model % t2(prof)
  profiles(i) % s2m % q     = max(model % q2(prof), qmin)
  profiles(i) % s2m % p     = model % pstar(prof) * 0.01
  profiles(i) % s2m % u     = model % u10(prof)
  profiles(i) % s2m % v     = model % v10(prof)
  profiles(i) % s2m % wfetc = 100000.0

  ! Ozone (s2m % o) ignored as it isn't used in RTTOV

  !--------------------------------------
  ! 2.3 Miscellaneous single-level fields
  !--------------------------------------

  ! Geolocation

  if ( associated(obs % zsurf) ) then
    zsurf = obs % zsurf(prof)
  else
    zsurf = model % zsurf(prof)
  end if
  profiles(i) % elevation = zsurf / 1000.0
  profiles(i) % latitude  = model % grid % lat(prof)
  if ( model % grid % lon(prof) < 0.0 ) then
    profiles(i) % longitude = model % grid % lon(prof) + 360.0
  else
    profiles(i) % longitude = model % grid % lon(prof)
  end if

  if ( obs_grid ) then
    if ( associated(obs % satzen)  ) profiles(i) % zenangle = obs % satzen(prof)
    if ( associated(obs % satazim) ) profiles(i) % azangle  = obs % satazim(prof)
    if ( associated(obs % solzen)  ) profiles(i) % sunzenangle = obs % solzen(prof)
    if ( associated(obs % solazim) ) profiles(i) % sunazangle  = obs % solazim(prof)
  else
    profiles(i) % zenangle    = zen_angle
    profiles(i) % azangle     = 0.0
    profiles(i) % sunzenangle = 0.0
    profiles(i) % sunazangle  = 0.0
  end if

  ! Cloud parameters

  profiles(i) % ctp = 850.0
  profiles(i) % cfraction = 0.0
  if ( ir_addclouds ) then
    profiles(i) % idg = ircloud_idg
    profiles(i) % ice_scheme = ircloud_ice_scheme
  end if

end do

!-------------------------------------------------------------------------------
! 3. Atmosphere variables
!-------------------------------------------------------------------------------

! nlevels and nlayers already set when initialised

! Use the first profile to determine direction of the model z-axis. In RTTOV
! the 1st level is at the top of the atmosphere (-ve z axis) but input profiles
! are often the other way round so we need to check whether they should be
! inverted.

z_direction = sign(1.0_8, model % p(prof_list(1),1)-model % p(prof_list(1),2))
if ( z_direction < 0 ) then
  level1 = -1
  layer1 = -1
else
  level1 = nlevels
  layer1 = nlevels-1
end if

!-------------------------
! 3.1 Compulsory variables
!-------------------------

! Pressure (hPa from Pa)

do j = 1, nlevels
  do i = 1, nprofs
    prof = prof_list(i)
    profiles(i) % p(abs(level1-j+1)) = model % p(prof,j)*0.01
  end do
end do

! Temperature (K)

do j = 1, nlevels
  do i = 1, nprofs
    prof = prof_list(i)
    profiles(i) % t(abs(level1-j+1)) = model % t(prof,j)
  end do
end do

! Humidity (kg/kg)
! Note that 'specific' values are the mass ratio of the gas to the total mass
! of air including that gas i.e., moist air. This is gas units option 1 in
! RTTOV. Note that ALL GASES MUST BE IN THE SAME UNITS.

! Some models occasionally give negative values for gas concentrations
! so ensure these are not passed into RTTOV.

do j = 1, nlevels
  do i = 1, nprofs
    prof = prof_list(i)
    profiles(i) % q(abs(level1-j+1)) = max(model % q(prof,j), qmin)
    profiles(i) % gas_units = 1
  end do
end do

!-------------------
! 3.2 Optional gases
!-------------------

! Ozone (kg/kg)

if ( associated(model % o3) .and. ozone_data ) then
  do j = 1, nlevels
    do i = 1, nprofs
      prof = prof_list(i)
      profiles(i) % o3(abs(level1-j+1)) = max(model % o3(prof,j), o3min)
    end do
  end do
end if

! co2, n2o, co, ch4, so2 currently ignored

!-------------
! 3.3 Aerosols
!-------------

! Currently ignored

!-----------
! 3.4 Clouds
!-----------

! See also single-level variables above to set size distribution and crystal shape

if ( .not. run_scatt ) then

  !--------------
  ! 3.4.1 MW only
  !--------------

  if ( associated(model % clw) .and. clw_data ) then
    do j = 1, nlevels
      do i = 1, nprofs
        prof = prof_list(i)
        profiles(i) % clw(abs(level1-j+1)) = model % clw(prof,j)
      end do
    end do
  else
    if ( output_mode >= output_debug ) then
      print '(a)', 'Running MW calculations without CLW data'
    end if
  end if

  !--------------
  ! 3.4.2 IR only
  !--------------

  if ( ir_addclouds ) then

    ! Make sure we always use the bottom level as the top level never has
    ! cloud in it.

    if ( z_direction < 0 ) then
      offset = 1
    else
      offset = 0
    end if

    ! Total cloud fraction

    if ( associated(model % cfrac) ) then
      do j = 1, nlevels-1
        do i = 1, nprofs
          prof = prof_list(i)
          profiles(i) % cfrac(abs(layer1-j+1)) = model % cfrac(prof,j+offset)
        end do
      end do
    end if

    ! Cloud amounts (kg/kg)

    ! Note: we assume that the model inputs are layer means and as such the cloud
    !       fraction is used to scale the value. Units should be kg/kg.
    ! RTTOV defines 6 types of cloud. These are
    ! 1 = Stratiform continental
    ! 2 = Stratiform maritime
    ! 3 = Cumulus continental clean
    ! 4 = Cumulus continental polluted
    ! 5 = Cumulus maritime
    ! 6 = Cirrus

    profiles(:) % mmr_cldaer = .true.   ! Units kg/kg

    ! Stratiform liquid cloud

    if ( associated(model % clw) ) then
      do j = 1, nlevels-1
        do i = 1, nprofs
          prof = prof_list(i)
          cloud = model % clw(prof,j+offset)
          cfrac = 0.
          if ( associated(model % cfrac_liq) ) then
            cfrac = model % cfrac_liq(prof,j+offset)
          else if ( associated(model % cfrac) ) then
            cfrac = model % cfrac(prof,j+offset)
          end if
          if ( cfrac > 0. ) cloud = cloud / cfrac
          if ( profiles(i) % skin % surftype == surftype_land ) then
            profiles(i) % cloud(1,abs(layer1-j+1)) = cloud
            profiles(i) % cloud(2,abs(layer1-j+1)) = 0.
          else
            profiles(i) % cloud(2,abs(layer1-j+1)) = cloud
            profiles(i) % cloud(1,abs(layer1-j+1)) = 0.
          end if
        end do
      end do
    end if

    ! Stratiform ice cloud

    if ( associated(model % ciw) ) then
      do j = 1, nlevels-1
        do i = 1, nprofs
          prof = prof_list(i)
          cloud = model % ciw(prof,j+offset)
          cfrac = 0.
          if ( associated(model % cfrac_ice) ) then
            cfrac = model % cfrac_ice(prof,j+offset)
          else if ( associated(model % cfrac) ) then
            cfrac = model % cfrac(prof,j+offset)
          end if
          if ( cfrac > 0. ) cloud = cloud / cfrac
          profiles(i) % cloud(6,abs(layer1-j+1)) = cloud
        end do
      end do
    end if

    ! Convective cloud is not output by UM GA6 or IFS
    ! This code needs modifying for UM GA7 (not yet operational)

!     if ( associated(model % conv_cloud) ) then
!       do j = 1, nlevels-1
!         do i = 1, nprofs
!           prof = prof_list(i)
!           cloud = model % conv_cloud(prof,j+offset)
!           if ( associated(model % cfrac_conv) ) then
!             cloud = cloud / model % cfrac_conv(prof,j+offset)
!           end if
!           if ( profiles(i) % skin % surftype == surftype_land ) then
!             profiles(i) % cloud(3,abs(layer1-j+1)) = cloud
!             profiles(i) % cloud(5,abs(layer1-j+1)) = 0.
!           else
!             profiles(i) % cloud(5,abs(layer1-j+1)) = cloud
!             profiles(i) % cloud(3,abs(layer1-j+1)) = 0.
!           end if
!         end do
!       end do
!     end if

  end if

else

  !-----------------------------
  ! 3.4.3 MW scattering profiles
  !-----------------------------

  do i = 1, nprofs
    cld_profiles(i) % use_totalice = use_totalice
    cld_profiles(i) % mmr_snowrain = mmr_snowrain
  end do

  ! half-level pressures. In RTTOV, there are more half-levels than levels.

  nhalflevels = size(model % ph, dim=2)
  if ( z_direction < 0 ) then
    halflevel1 = -1
  else
    halflevel1 = nlevels+1
  end if

  fill_top = .false.
  if ( nhalflevels > nlevels ) then
    offset = 1
  else if ( nhalflevels == nlevels ) then
    if ( minval(model % ph(prof_list(1),:)) > minval(model % p(prof_list(1),:)) ) then
      offset = 1
      fill_top = .true.
    else
      offset = 0
    end if
  else
    offset = 0
    fill_top = .true.
  end if

  do j = 1, nhalflevels
    do i = 1, nprofs
      prof = prof_list(i)
      cld_profiles(i) % ph(abs(halflevel1-j+offset)) = model % ph(prof,j)*0.01
    end do
  end do

  ! Lowest half-level pressure must be the same as the surface. If the top level
  ! isn't set then fill that with a small value.

  do i = 1, nprofs
    prof = prof_list(i)
    cld_profiles(i) % ph(nlevels+1) = profiles(i) % s2m % p
    if ( fill_top ) cld_profiles(i) % ph(1) = 0.001
  end do

  ! Cloud cover (area fraction)

  do j = 1, nlevels
    do i = 1, nprofs
      prof = prof_list(i)
      cld_profiles(i) % cc(abs(level1-j+1)) = model % cfrac(prof,j)
    end do
  end do

  ! CLW (kg/kg)

  do j = 1, nlevels
    do i = 1, nprofs
      prof = prof_list(i)
      cld_profiles(i) % clw(abs(level1-j+1)) = model % clw(prof,j)
    end do
  end do

  ! CIW (kg/kg)

  if ( use_totalice ) then

    do j = 1, nlevels
      do i = 1, nprofs
        prof = prof_list(i)
        cld_profiles(i) % totalice(abs(level1-j+1)) = model % ciw(prof,j)
      end do
    end do

  else

    do j = 1, nlevels
      do i = 1, nprofs
        prof = prof_list(i)
        cld_profiles(i) % ciw(abs(level1-j+1)) = model % ciw(prof,j)
      end do
    end do

  end if

  ! Rain (kg/kg or kg/m^2/s according to mmr_snowrain); not compulsory

  if ( associated(model % rain) ) then
    do j = 1, nlevels
      do i = 1, nprofs
        prof = prof_list(i)
        cld_profiles(i) % rain(abs(level1-j+1)) = model % rain(prof,j)
      end do
    end do
  else
    print '(a)', 'Running scattering code without rain data'
  end if

  ! Solid precipitation (kg/kg or kg/m^2/s according to mmr_snowrain); not compulsory

  if ( associated(model % snow) .and. .not. use_totalice ) then
    do j = 1, nlevels
      do i = 1, nprofs
        prof = prof_list(i)
        cld_profiles(i) % sp(abs(level1-j+1)) = model % snow(prof,j)
      end do
    end do
  end if

end if

end subroutine radsim_model_to_rttov
